
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <config.h>
#include "gjs-enums.h"
#include "libgjs-private/gjs-util.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gjs-util.h" */

GType
gjs_locale_category_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GJS_LOCALE_CATEGORY_ALL), "GJS_LOCALE_CATEGORY_ALL", "all" },
        { C_ENUM (GJS_LOCALE_CATEGORY_COLLATE), "GJS_LOCALE_CATEGORY_COLLATE", "collate" },
        { C_ENUM (GJS_LOCALE_CATEGORY_CTYPE), "GJS_LOCALE_CATEGORY_CTYPE", "ctype" },
        { C_ENUM (GJS_LOCALE_CATEGORY_MESSAGES), "GJS_LOCALE_CATEGORY_MESSAGES", "messages" },
        { C_ENUM (GJS_LOCALE_CATEGORY_MONETARY), "GJS_LOCALE_CATEGORY_MONETARY", "monetary" },
        { C_ENUM (GJS_LOCALE_CATEGORY_NUMERIC), "GJS_LOCALE_CATEGORY_NUMERIC", "numeric" },
        { C_ENUM (GJS_LOCALE_CATEGORY_TIME), "GJS_LOCALE_CATEGORY_TIME", "time" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GjsLocaleCategory"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

